/**
Author:  Luis Gerardo de la Fraga
Scubes is a program to visualize vox-solids
Copyright (C) 2001 Luis Gerardo de la Fraga

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's e-mail: fraga@cs.cinvestav.mx
Address: 
Seccin de Computacin. Dpto. de Ingeniera Elctrica
CINVESTAV-IPN                       
Av. Instituto Politcnico Nacional No. 2508
07300 Mxico, D.F. 
Mxico
**/

/**
	Definitions for 4cubes program
	Fraga 28/11/2000
**/
#ifndef SPACE3D_H
#define SPACE3D_H

#include <qwidget.h>
#include <qgl.h>
#include <qtimer.h>
#include <qlist.h>


class Sphere
{
public:
	enum Type { POLYGON, WIRED };
	typedef struct {
		int n;
		int color[3];
		float norm[3];
		float *v;
	} Plane;

	Sphere ( Type a ) { 
		// allPlanes=NULL; 
		nplanes=0; greytone=255; typecolor=0; 
		wtype = a;
	}
	~Sphere ( );
	void PaintSphere( void );
	int AllocatePlanes( int m );
	void PutColor ( int r, int g, int b ) {
		front_mat_diffuse[0] = r/255.0;
		front_mat_diffuse[1] = g/255.0;
		front_mat_diffuse[2] = b/255.0;
		front_mat_diffuse[3] = 1.0;
	}
	void PutGreyToneValue ( GLubyte g ) {
		greytone = g;
	}
	void PutTypeColor ( int val ) {
		typecolor = val;
	}
	void PutVertexValues( int j, int i, float x, float y, float z );
	void PutMyNormal( int h, float nx, float ny, float ny );
	int Create_Polygon( int i, int m );
private:

	Plane **allPlanes;
	int nplanes;
	Type wtype;
	float front_mat_diffuse[4];
	int typecolor; //0 = Color, 1 = greytones
	GLubyte greytone;
};

class Space3D : public QGLWidget
{
	Q_OBJECT  // must include this if you use Qt signals/slots
public:
	Space3D( const char *fname, QWidget *parent=0, const char *name=0 );
	QSizePolicy sizePolicy() const;
	void getPosition ( int *x, int *y ) { *x = xmouse; *y = ymouse; }
	void getSize ( int *w, int *h ) { 
		*w = windowWidth; *h = windowHeight; 
	}
	void save_canvas( void );
	enum typeAngle { THETA, PHI, PSI };

public slots:
	void timerDone( void );
	void selectColor( int val ){
		typecolor = val;
		Sphere *phe;
        for ( phe=listspheres.first(); phe != NULL; phe=listspheres.next() )
			phe -> PutTypeColor( typecolor );
		updateGL( );
	}
	void greyValue( int val ){
		greytone = 0x00ff & val;
		Sphere *phe;
        for ( phe=listspheres.first(); phe != NULL; phe=listspheres.next() )
			phe -> PutGreyToneValue( greytone );
		updateGL( );
	}
	void angValue1( int val ){
		theta = val; 
		updateGL( );
	}
	void angValue2( int val ){
		phi = val; 
		updateGL( );
	}
	void selectLight( int val ){
		if ( !val ) 
			glEnable( GL_LIGHTING );
		else
			glDisable( GL_LIGHTING );

		updateGL( );
	}
	void selLightSubject( int val ) {
		wholight = val;
		updateGL( );
	}
/** Not used because of we use only one color per polygoN
	void selectShade( int val ){
		if ( !val ) 
			glShadeModel( GL_SMOOTH );
		else
			glShadeModel( GL_FLAT );

		updateGL( );
	}
**/

	void selectWhichAng( int val ) 
	{
		switch( val ){
			case 0:
				wangle = THETA;
				break;
			case 1:
				wangle = PHI;
				break;
			case 2:
				wangle = PSI;
				break;
		}
	}
	void angAniVal1( int ang ){ start_ang = ang; }
	void angAniVal2( int ang ){ final_ang = ang; }
	void animbutton(int state);
	void numImages(int n){  nimages=n; }

	void dist_slider( int value ) {
		scale = 2*value;
		
		// scale = m_val*(distance-1) + 1;
		updateGL( );
	}

signals:
	void changeButton( ); // called when animation stops
 
protected:
	void initializeGL( void );
	void resizeGL( int w, int h );
	void paintGL( void );
	void mouseMoveEvent( QMouseEvent * );
	// void mousePressEvent( QMouseEvent * );
 
private:
	QList<Sphere> listspheres;

	int construct_spheres( const char *filename );
	void InitMaterials(void);

	QTimer *timer;

	int windowWidth, windowHeight;   // canvas width and height

	int nspheres;  // the number of vox-solids
	int dim;         // Dimension of total show scene: dim x dim x dim

	int typecolor; //0 = Color, 1 = greytones
	int wholight;  // 0 light fixed to viewer, 1 fixed to object
	GLubyte greytone;

	float yrot, xrot;  // Encene rotation
	float theta, phi;
	int xmouse, ymouse;  // Mouse pointer coordinates

	// For animation (save images screen)
	int counter;
	int nimages;
	typeAngle wangle;
	float final_ang, start_ang, initial_ang, delta_ang;
	float distance, scale, m_val;
};

#endif //  SPACE3d_H
